unit NewDBGrid;

{$ifdef Ver90} { Delphi 2.0x }
  {$define DelphiLessThan4}
{$endif}
{$ifdef Ver93} { C++ Builder 1.0x }
  {$define DelphiLessThan4}
{$endif}
{$ifdef Ver100} { Delphi 3.0x }
  {$define DelphiLessThan4}
{$endif}
{$ifdef Ver110} { C++ Builder 3.0x }
  {$define DelphiLessThan4}
{$endif}

{$ifdef DelphiLessThan4}
  'This class requires Delphi 4 or later'
{$endif}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, DBGrids;

type
  TResizeEvent = procedure(Sender: TCustomDBGrid; Column: TColumn;
    Width: Cardinal) of object;

  TNewDBGrid = class(TDBGrid)
  private
    //Column being resized
    FColumn: TColumn;
    //Horizontal position of mouse when resize started
    FOldX: Integer;
    //Private fields for event handlers
    FOnResized,
    FOnResizing: TResizeEvent;
  protected
    //Generic resize event code
    procedure DoResize(NewX: Integer; ResizeEvent: TResizeEvent); virtual;
    //Routines to trap resize operations
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  published
    //New events
    property OnResizing: TResizeEvent read FOnResizing write FOnResizing;
    property OnResized:  TResizeEvent read FOnResized  write FOnResized;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Clinic', [TNewDBGrid]);
end;

{ TNewDBGrid }

procedure TNewDBGrid.DoResize(NewX: Integer; ResizeEvent: TResizeEvent);
var
  Width: Integer;
begin
  if (FGridState = gsColSizing) and
     Assigned(FColumn) and
     Assigned(ResizeEvent) then
  begin
    //Calculate new column width
    Width := FColumn.Width + NewX - FOldX;
    //Deal with silly resize requests
    if Width <= 1 then
      Width := FColumn.Width;
    //Trigger specified event handler
    ResizeEvent(Self, FColumn, Width)
  end
end;

procedure TNewDBGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  ColumnIndex: Integer;
  Pos, Ofs: Integer;
  State: TGridState;
  DrawInfo: TGridDrawInfo;
begin
  inherited;
  if FGridState = gsColSizing then
  begin
    //Set up TGridDrawInfo record for next statement
    CalcDrawInfo(DrawInfo);
    //Ask grid which cell is being resized
    CalcSizingState(X, Y, State, ColumnIndex, Pos, Ofs, DrawInfo);
    //Take indicator column into account
    ColumnIndex := RawToDataColumn(ColumnIndex);
    if ColumnIndex >= 0 then
      FColumn := Columns[ColumnIndex];
    FOldX := X
  end
  else
    FColumn := nil
end;

procedure TNewDBGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  //Possible resizing occurring
  DoResize(X, FOnResizing)
end;

procedure TNewDBGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  //Possible resize completed
  DoResize(X, FOnResized);
  //Finished resizing the column, so clear it
  FColumn := nil;
  inherited;
end;

end.